import {PostToConnectionCommand} from '@aws-sdk/client-apigatewaymanagementapi';
import {AnyTRPCRouter, callTRPCProcedure, inferRouterContext} from '@trpc/server';
import {isObservable, Observable} from '@trpc/server/observable';
import {isAsyncIterable} from '@trpc/server/unstable-core-do-not-import';
import {APIGatewayProxyEvent, APIGatewayProxyResult} from 'aws-lambda';
import {getApiGatewayManagementApiClient} from '../getApiGatewayManagementApiClient';
import {getConnection} from '../getConnection';
import {TrpcSubscriptionRequest} from './TrpcSubscriptionRequest';
import {TrpcSubscriptionResponse} from './TrpcSubscriptionResponse';

export const awsLambdaTrpcSubscriptionRequestHandler = <Router extends AnyTRPCRouter>(opts: {
  router: Router
  createContext?: (event: APIGatewayProxyEvent) => inferRouterContext<Router> | Promise<inferRouterContext<Router>>
}) => {
  const {router, createContext} = opts;

  return async (event: APIGatewayProxyEvent): Promise<APIGatewayProxyResult> => {
    const connection = getConnection(event);

    if (!event.body || event.body === '[]')
      return {statusCode: 200, body: '{}'};

    const request = (JSON.parse(event.body) as TrpcSubscriptionRequest);

    const sendMessage = async (response: TrpcSubscriptionResponse) => {
      await getApiGatewayManagementApiClient(connection.endpoint).send(
          new PostToConnectionCommand({
            ConnectionId: connection.connectionId,
            Data: JSON.stringify(response),
          }),
      );
    };
    const stopConnection = async (): Promise<void> => {
      await sendMessage({
        id: request.id,
        result: {type: 'stopped'},
      });
    };

    try {
      if (request.method === 'subscription.stop') {
        await stopConnection();
      }

      if (request.method === 'subscription') {
        await sendMessage({
          id: request.id,
          result: {type: 'started'},
        });

        const procedure = (await callTRPCProcedure({
          procedures: router._def.procedures,
          path: request.params.path,
          type: 'subscription',
          ctx: await createContext?.(event) || {},
          getRawInput: async () => request.params.input,
        }));

        if (isAsyncIterable(procedure)) {
          await handleAsyncGenerator(procedure, request, sendMessage, stopConnection);
        } else if (isObservable(procedure)) {
          await handleObservable(procedure, request, sendMessage);
        } else {
          throw 'Unable to determine subscription type';
        }
      }
    } catch (err) {
      console.error(err);
      await stopConnection();
    }

    return {
      statusCode: 200,
      body: '{}',
    };
  };
};


const handleObservable = async (
    observable: Observable<unknown, unknown>,
    request: TrpcSubscriptionRequest,
    sendMessage: (message: TrpcSubscriptionResponse) => Promise<void>,
) => {
  const messageQueue: TrpcSubscriptionResponse[] = [];
  observable.subscribe({
    next: (data: unknown) => {
      messageQueue.push({
        id: request.id,
        result: {
          type: 'data',
          data: data,
        },
      });
    },
    error: (err) => {
      console.error(err);
      messageQueue.push({
        id: request.id,
        result: {type: 'stopped'},
      });
    },
    complete: () => {
      messageQueue.push({
        id: request.id,
        result: {type: 'stopped'},
      });
    },
  });

  let stopped = false;
  while (!stopped) {
    const message = messageQueue.shift();
    if (!message)
      continue;

    await sendMessage(message);

    stopped = message.result.type === 'stopped';
  }
};

const handleAsyncGenerator = async (
    generator: AsyncIterable<unknown>,
    request: TrpcSubscriptionRequest,
    sendMessage: (message: TrpcSubscriptionResponse) => Promise<void>,
    stopConnection: () => Promise<void>
) => {
  try {
    for await (const data of generator) {
      await sendMessage({
        id: request.id,
        result: {
          type: 'data',
          data: data,
        },
      });
    }
  } catch (err) {
    console.error(err);
  } finally {
    await stopConnection();
  }
};
